% lab09_ex_filtering.m  - FIR digital filtering example
clear all; close all;

% Parameters
task=4;                  % 1=LP, 2=HP, 3=BP, 4=BS - filter type
fs=2000;                 % sampling ratio
f0=100; f1=200; f2=300;  % cut-off frequencies
L=100;                   % half of the filter length
N=2*L+1                  % filter length: 2*L or 2*L+1
M=(N-1)/2;               % central point of weigts

% Signal
Nx=2000; n=0:Nx-1; dt=1/fs; t=dt*n;
x1 = cos(2*pi*10*t); x2 = cos(2*pi*250*t); % two cosines 
x  = x1 + x2;
if(task==1 | task==4) xref = x1; else xref = x2; end

% Matlab functions - design of filter impulse responses
h1=fir1(N-1,f0/(fs/2),'low');            % low-pass
h2=fir1(N-1,f0/(fs/2),'high');           % high-pass
h3=fir1(N-1,[f1,f2]/(fs/2),'bandpass');  % band-pass
h4=fir1(N-1,[f1,f2]/(fs/2),'stop');      % band-stop
if    (task==1) h=h1;
elseif(task==2) h=h2;
elseif(task==3) h=h3;
else            h=h4;
end
figure; stem(h); xlabel('n'); title('h(n)'); grid; pause

% Roots of filter transfer function
z = roots(h); z = z( find( abs(z) < 10 ) );
alpha = 0 : pi/1000 : 2*pi; co=cos(alpha); si=sin(alpha);
figure; plot(co,si,'k-',real(z),imag(z),'bo'); title('TF Zeros'); grid; pause

% Filter frequency response
f = 0 : fs/2000 : fs/2;
z = exp(-j*2*pi*f/fs);        % Z transform z^(-1)
H = polyval( fliplr(h), z );  % H = freqz(h,1,f,fs); 
figure
subplot(211); plot(f,20*log10(abs(H))); xlabel('f (Hz)'); title('|H(f)|'); grid;
subplot(212); plot(f,unwrap(angle(H))); xlabel('f (Hz)'); title('Phase H(f)'); grid; pause

% FIR digital filtering
%y = conv(x,h); y = filter(h,1,x);
N = length(h);
bx = zeros(1,N);
for n = 1 : Nx
    bx = [ x(n) bx(1:N-1) ];
    y(n) = sum( bx .* h );
end

% Figures
n=0:Nx-1; 
figure; % observe that first N-1 samples of y(n) are erroneous (transient)
subplot(211); plot(n,x,'.-'); xlabel('n'); title('x(n)'); grid;
subplot(212); plot(n,y,'.-'); xlabel('n'); title('y(n)'); grid; pause
figure; nx = L+1: Nx-L; ny = N:Nx;
subplot(111); plot(t(nx),xref(nx),'bo-',t(nx),y(ny),'r*-'); xlabel('n');
title('x(n) BLUE (o),   y(n) RED (*)'); grid; pause
figure; % signal spectra of the second halves of samples (transients out!)
k=Nx/2+1:Nx;  f0 = fs/(Nx/2);  f=f0*(0:Nx/2-1); 
subplot(211); plot(f,20*log10(abs(2*fft(x(k)))/(Nx/2))); title('|X(f)|'); grid;
subplot(212); plot(f,20*log10(abs(2*fft(y(k)))/(Nx/2))); title('|Y(f)|'); grid;
xlabel('f (Hz)'); pause
