
% lab09_ex_filters_all.m

% Lab 9. DIGITAL FIR FILTERS - QUICK INTRODUCTION
%
% SUMMARY
% Digital FIR filter is defined by the following equation:
%            M
%    y(n) = SUM( b(m)*x(n-m) ),   y(n) = b0*x(n-0) + b1*x(n-1) + ... + bM*x(n-M))
%           m=0
%  Filter output y(n) is a weighted sum (with weights b()) of some last input samples x().
%
% Design of weights { b(m) } is called FIR FILTER DESIGN.
% Calculation of y(n) having x(n) is called FIR DIGITAL FILTERING.
% FIR filters require many (hundreds of) weights to have sharp frequency response.
% So they are computationally consuming.
% However they are always stable and not distorting the signal shape.
% If properly designed ((a)symmetry of weights) they have always linear phase response.
% Due to this they are only delaying all signal components, which are passing through them.
% The delay is equal half of the filter length (half of number of weights bm). 
%
% In Matlab coefficients { b(m) } can be designed using many methods and functions:
% fir1(), fir2(), firls(), firpm(), fircls(), cfirpm()
% We will become familiar with first three methods.
% In this programs we will concentrate on designing the filter weights.
% After that we put the developed code into functions MYfir1(), MYfir2(), MYfirls().
% These functions can be called in general digital filtering program:
% Lab7_digital_filtering.m

% IMPORTANT!
% In this program weights {bm} are denoted as { h[m] } since for FIR filters
% wights {bm} are equal to samples of the filter impulse response.
% In this program weights {b(m)} are denoted as {h(m)}
% since for FIR filters b(m) represent samples of the filter impulse response 

clear all; close all;

% Filter parameters

fs = 1000;       % sampling frequency
f0 = 100;        % cut-off frequency for low-pass and high-pass filters
f1 = 200;        % low frequency for band-pass and band-stop filters
f2 = 300;        % high frequency for band-pass and band-stop filters
M = 100;         % half of the filter weights
N = 2*M + 1;     % number of weights {b(n)} - always odd
n = -M :1: M;    % non-causal weigths indexes

% ###########################################
% MYfir1 - window method: analitic solution found by the inverse DtFT
hALL = zeros(1,N); hALL(M+1)=1;                         % AllPass
hLP = sin(2*pi*f0/fs*n)./(pi*n); hLP(M+1) = 2*f0/fs;    % LowPass f0
hHP = hALL - hLP;                                       % HighPass
hLP1 = sin(2*pi*f1/fs*n)./(pi*n); hLP1(M+1) = 2*f1/fs;  % LowPass f1
hLP2 = sin(2*pi*f2/fs*n)./(pi*n); hLP2(M+1) = 2*f2/fs;  % LowPass f2
hBP = hLP2 - hLP1;                                      % BandPass [f1,f2]
hBS = hALL - hBP;                                       % BandStop [f1,f2]
hH = (1-cos(pi*n))./(pi*n); hH(M+1)=0;  % Hilbert filter -90deg phase shifter
hD = cos(pi*n)./n; hD(M+1)=0;           % differentiation filter
h1 = hLP;

% ###########################################
% MYfir2 - sampling in frequency method: inverse DFT of the requested freq response
NN = 4*N;
df=fs/NN; f=0:df:fs-df;
%indLP = find( f<f0 | f>(fs-f0) );   % Low-Pass
indLPdn = find(f<f0); indLPup = find(f>(fs-f0)); indLP = [ indLPdn indLPup ];
indHP = find( f>f0 & f<(fs-f0) );   % High-Pass

indLP1 = find( f<f1 );  % Low-Pass f1
indLP2 = find( f<f2 );  % Low-Pass f2
indBP = find( indLP2 > indLP1(end) );
indBP = [ indBP NN-indBP+2 ];

H = zeros(1,NN);
HLP = H; HLP(indLP) = ones(1,length(indLP)); HLP(indLPdn(end))=0.5; HLP(indLPup(1))=0.5;
hLP = ifft( HLP ); % hLP=fftshift(hLP);
hLP = [ hLP(M+1:-1:2) hLP(1:M+1) ];

HHP = H; HHP(indHP) = ones(1,length(indHP));  HHP(indHP(1))=0.5; HHP(indHP(end))=0.5; 
hHP = ifft( HHP ); hHP=fftshift(hHP);

HBP = H; HBP(indBP) = ones(1,length(indBP));
hBP = ifft( HBP ); hBP=fftshift(hBP);

h2 = hLP;

% Design Band-Stop Filter h2 = hBS;

% ###########################################
% MYfirls - least-squares method
% Specify your requirements: 
% P points of frequency response Ad() for radial frequencies 2*pi*p/P, p=0,1,2,...,P-1
  Q = M;              % half of the filter length
  P = 10*2*Q;            % number of points (even; P=>N)
  L1 = floor(f0/fs*P)+40, pause  % number of first ones (then transient points, finally zeros)
 % Ap = [ ones(1,L1) 0.75 0.25 zeros(1,P-(2*L1-1)-4) 0.25 0.75 ones(1,L1-1)]; 
   Ap = [ ones(1,L1) 0.5 zeros(1,P-(2*L1-1)-2) 0.5 ones(1,L1-1)]; 
  Ap = Ap'; 
% Chose weighting coefficients w(p), p=0,1,2,...,P-1, for Pass/Trans/Stop regions
  wp = 1;         %  weight for PassBand
  wt = 1;         %  weight for TransientBand
  ws = 10000000;  %  weight for StopBand
% w = [ wp*ones(1,L1) wt wt ws*ones(1,P-(2*L1-1)-4) wt wt wp*ones(1,L1-1) ]; 
  w = [ wp*ones(1,L1) wt ws*ones(1,P-(2*L1-1)-2) wt wp*ones(1,L1-1) ]; 
  W = zeros(P,P); % 
  for p=1:P       %  matrix with weights on the main diagonal
     W(p,p)=w(p); %
  end             %
% Find matrix F being solution of the matrix equation  W*F*h = W*(Ap + err) 
  F = [];
  n = 0 : Q-1; 
  for p = 0 : P-1
     F = [ F; 2*cos(2*pi*(Q-n)*p/P) 1 ]; 
  end
% Find h[n] minimizing error of W*F*h = W*Ap
% h = pinv(W*F)*(W*Ap);  % method #1
  h = (W*F)\(W*Ap);      % method #2
  h3 = [ h; h(Q:-1:1) ]';
  
% Rysunki
  n = 0 : 2*Q; subplot(111); stem(n,h3); grid; 
  title('Impulse respone h3[n]'); xlabel('sample number'); pause
  NF=500; wn =0:pi/(NF-1):pi; fn = wn/(2*pi); H = freqz(h3,1,wn); 
  subplot(211); plot(fn,abs(H)); grid; title('Magnitude of freq response');
  subplot(212); plot(fn,180/pi*unwrap(angle(H))); grid; 
  title('Phase of freq response'); ylabel('[grad]'); xlabel('f norm [Hz]'); pause
  subplot(111); plot(fn,20*log10(abs(H))); grid; title('Magnitude of freq response');
  xlabel('f norm [Hz]'); axis([0 0.5 -120 10]); pause

% ###########################################
% Windowing
%w = chebwin(N,100)';   % Asl=60, 80, 100, 120 attenuation of side-lobes in stop-band
 w = kaiser(N,10)';     % beta=7, 10, 0-15
h1 = h1.*w;
h2 = h2.*w;
%h3 = h3.*w;
figure;
subplot(311); stem(n,h1); title('h1[n]'); grid;
subplot(312); stem(n,h2); title('h2[n]'); grid;
subplot(313); stem(n,h3); title('h3[n]'); grid; pause

% ###########################################
%Freq response H(f) = H(z) for z=...
f = 0:0.25:fs/2; z = exp(-j*2*pi*f/fs);
H1 = polyval(h1,z);
H2 = polyval(h2,z);
H3 = polyval(h3,z);
figure; plot(f,20*log10(abs(H1)),'b-',f,20*log10(abs(H2)),'r-',f,20*log10(abs(H3)),'k-');
grid; xlabel('f [Hz]'); title('Abs( H(f) )'); pause
figure; plot(f,unwrap(angle(H1)),'b-',f,unwrap(angle(H2)),'r-',f,unwrap(angle(H3)),'k-');
grid; xlabel('f [Hz]'); title('Phase( H(f) )'); pause

% #######
% TESTING
% #######

h = h1;

% Signal
Nx = 1000;
dt = 1/fs; t = dt*(0:Nx-1);
f1 = 10; f2=250;
x = 2*cos(2*pi*f1*t) + 2*cos(2*pi*f2*t);
figure; plot(t,x,'b-'); xlabel('t [s]'); grid; pause

% Filternig
% y = filter(h,1,x);    % Matlab filtering function for IIR fiters with {b=h,a=1} coeffs
% y = conv(x,h);        % Matlab filtering function for FIR filters with b=h coeffs only 
N = length(h);
bx = zeros(1,N);
for n=1:Nx
    bx = [ x(n) bx(1:N-1)];  % bx(2:N)=bx(1:N-1); bx(1)=x(n);
    y(n) = sum( bx .* h);    % weighting summation of input
end

figure; % observe that first N-1 samples of y(n) are errorness (an initial transient)
subplot(211); plot(t,x,'b-'); xlabel('t [s]'); title('IN  x(t)'); grid;
subplot(212); plot(t,y,'b-'); xlabel('t [s]'); title('OUT y(t)'); grid; pause

figure; % only second half of the signal without initial transient
nn = Nx/2+1:Nx; f0=fs/(Nx/2); f=f0*(0:Nx/2-1); w=chebwin(Nx/2,100)'; w=w/sum(w);
subplot(211); plot(f,20*log10(abs(fft(w.*x(nn))))); grid; xlabel('f [Hz]'); title('IN  |X(f)| [dB]');
subplot(212); plot(f,20*log10(abs(fft(w.*y(nn))))); grid; xlabel('f [Hz]'); title('OUT |Y(f)| [dB]');
pause
