
% lab09_ex_filters_demo.m

% Lab 9. DIGITAL FIR FILTERS - QUICK INTRODUCTION
%
% SUMMARY
% Digital FIR filter is defined by the following equation:
%            M
%    y(n) = SUM( b(m)*x(n-m) ),   y(n) = b0*x(n-0) + b1*x(n-1) + ... + bM*x(n-M))
%           m=0
%  Filter output y(n) is a weighted sum (with weights b()) of some last input samples x().
%
% Design of weights { b(m) } is called FIR FILTER DESIGN.
% Calculation of y(n) having x(n) is called FIR DIGITAL FILTERING.
% FIR filters require many (hundreds of) weights to have sharp frequency response.
% So they are computationally consuming.
% However they are always stable and not distorting the signal shape.
% If properly designed ((a)symmetry of weights) they have always linear phase response.
% Due to this they are only delaying all signal components, which are passing through them.
% The delay is equal half of the filter length (half of number of weights bm). 
%
% In Matlab coefficients { b(m) } can be designed using many methods and functions:
% fir1(), fir2(), firls(), firpm(), fircls(), cfirpm()
% We will become familiar with first three methods.
% In this programs we will concentrate on designing the filter weights.
% After that we put the developed code into functions MYfir1(), MYfir2(), MYfirls().
% These functions can be called in general digital filtering program:
% lab08_digital_filtering.m

% IMPORTANT!
% In this program weights {bm} are denoted as { h[m] } since for FIR filters
% wights {bm} are equal to samples of the filter impulse response.
% In this program weights {b(m)} are denoted as {h(m)}
% since for FIR filters b(m) represent samples of the filter impulse response. 

clear all; close all;

% Filter parameters

fs = 1000;       % sampling frequency
f0 = 100;        % cut-off frequency for low-pass and high-pass filters
f1 = 200;        % low frequency for band-pass and band-stop filters
f2 = 300;        % high frequency for band-pass and band-stop filters
M = 100;         % half of the filter weights
N = 2*M + 1;     % number of weights {h(n)} - always odd
n = -M :1: M;    % non-causal weigths indexes

% ###########################################
% MYfir1 - window method: analitic solution found by the inverse DtFT
% try Matlab function fir1()
hALL = zeros(1,N); hALL(M+1)=1;                         % AllPass
hLP = sin(2*pi*f0/fs*n)./(pi*n); hLP(M+1) = 2*f0/fs;    % LowPass f0
hHP = hALL - hLP;                                       % HighPass
hLP1 = sin(2*pi*f1/fs*n)./(pi*n); hLP1(M+1) = 2*f1/fs;  % LowPass f1
hLP2 = sin(2*pi*f2/fs*n)./(pi*n); hLP2(M+1) = 2*f2/fs;  % LowPass f2
hBP = hLP2 - hLP1;                                      % BandPass [f1,f2]
hBS = hALL - hBP;                                       % BandStop [f1,f2]
hH = (1-cos(pi*n))./(pi*n); hH(M+1)=0;  % Hilbert filter -90deg phase shifter
hD = cos(pi*n)./n; hD(M+1)=0;           % differentiation filter
h1 = hLP;

% ###########################################
% MYfir2 - sampling in frequency method: inverse DFT of the requested freq response
% try Matlab function fir2()
K = 4;   % oversampling in frequency domain
NK = N*K; if(rem(NK,2)==1) NK=NK+1; end
df = fs/NK; f = df*(0 : NK/2);
H0 = zeros(1,NK/2+1);
H1 = ones(1,NK/2+1);

ind = find( f<=f0 );     % Low-Pass
HLP = H0; HLP(ind) = ones(1,length(ind)); HLP(ind(end))=0.5;

ind = find( f>=f0 );     % High-Pass
HHP = H0; HHP(ind) = ones(1,length(ind)); HLP(ind(1))=0.5;

ind1 = find( f< f1 );    % Band-Pass
ind2 = find( f<=f2 );    %
ind  = find( ind2 > ind1(end) );
HBP = H0; HBP(ind) = ones(1,length(ind));  HBP(ind(1))=0.5; HBP(ind(end))=0.5; 
HBS = H1; HBS(ind) = zeros(1,length(ind)); HBS(ind(1))=0.5; HBS(ind(end))=0.5; 

H = HLP;
H(NK:-1:NK/2+2) = H(2:NK/2);
h = ifft( H );
h2 = [ h(M+1:-1:2) h(1:M+1) ];

% Design Band-Stop Filter h2 = hBS;

% ###########################################
% MYfirls - least-squares method
% Try Matlab function firls()
% Specify your requirements: 
% P points of frequency response Ad() for pulsations 2*pi*p/P, p=0,1,...,P-1
% Example of low-pass filter design
  K = 4;                      % oversampling in frequency domain
  P = K*2*M;                  % number of points (even; P >= N=2M+1)
  L1 = floor(f0/fs*P), pause  % number of first ones
  Ad = [ ones(1,L1) 0.5 zeros(1,P-(2*L1-1)-2) 0.5 ones(1,L1-1)]; 
  Ad = Ad'; 
% Chose weighting coeffs w(p), p=0,1,2,...,P-1, for Pass/Trans/Stop regions
  wp = 1;         %  weight for PassBand
  wt = 1;         %  weight for TransientBand
  ws = 10000;     %  weight for StopBand
  w = [ wp*ones(1,L1) wt ws*ones(1,P-(2*L1-1)-2) wt wp*ones(1,L1-1) ]; 
  W = zeros(P,P); % 
  for p=1:P       %  matrix with weights on the main diagonal
     W(p,p)=w(p); %
  end             %
% Find matrix F being solution of the matrix equation  W*F*h = W*(Ad + err) 
  F = [];
  n = 0 : M-1; 
  for p = 0 : P-1
      F = [ F; 2*cos(2*pi*(M-n)*p/P) 1 ]; 
  end
% Find h[n] minimizing error of W*F*h = W*Ad
% h = pinv(W*F)*(W*Ad);  % method #1
  h = (W*F)\(W*Ad);      % method #2
  h3 = [ h; h(M:-1:1) ]';
  
% ###########################################
% Windowing
% w = chebwin(N,100)';  % Asl=60, 80, 100, 120 attenuation of side-lobes in stop-band
w = kaiser(N,10)';      % beta=7, 10, 0-15
h1 = h1.*w;
h2 = h2;    % h2.*w;    % try additional windowing 
h3 = h3;    % h3.*w;    % try additional windowing
figure; n = 0 : 2*M; 
subplot(311); stem(n,h1); title('h1(n) - window method'); grid;
subplot(312); stem(n,h2); title('h2(n) - invDFT method'); grid;
subplot(313); stem(n,h3); title('h3(n) - WLS method   '); grid; pause

% ###########################################
%Freq response H(f) = H(z) for z=...
f = 0:0.25:fs/2; z = exp(-j*2*pi*f/fs);  % Z transform z^(-1)
H1 = polyval( h1(end:-1:1), z );                 
H2 = polyval( h2(end:-1:1), z );
H3 = polyval( h3(end:-1:1), z );
text={'Window','InvDFT','WLS'};
figure; plot(f,20*log10(abs(H1)),'b-',f,20*log10(abs(H2)),'r-',f,20*log10(abs(H3)),'k-');
grid; xlabel('f [Hz]'); title('Abs( H(f) )'); legend(text); pause
figure; plot(f,unwrap(angle(H1)),'b-',f,unwrap(angle(H2)),'r-',f,unwrap(angle(H3)),'k-');
grid; xlabel('f [Hz]'); title('Phase( H(f) )'); legend(text); pause

% #######
% TESTING
% #######

h = h1;

% Signal
Nx = 1000;
dt = 1/fs; t = dt*(0:Nx-1);
f1 = 10; f2=250;
x = 2*cos(2*pi*f1*t) + 2*cos(2*pi*f2*t);  % two cosines
figure; plot(t,x,'b-'); xlabel('t [s]'); grid; pause

% Filternig
% y = filter(h,1,x);    % Matlab filtering function for IIR fiters with {b=h,a=1} coeffs
% y = conv(x,h);        % Matlab filtering function for FIR filters with b=h coeffs only 
N = length(h);
bx = zeros(1,N);
for n=1:Nx
    bx = [ x(n) bx(1:N-1)];  % bx(2:N)=bx(1:N-1); bx(1)=x(n);
    y(n) = sum( bx .* h);    % weighting summation of input
end

figure; % observe that first N-1 samples of y(n) are errorness (an initial transient)
subplot(211); plot(t,x,'b-'); xlabel('t [s]'); title('IN  x(t)'); grid;
subplot(212); plot(t,y,'b-'); xlabel('t [s]'); title('OUT y(t)'); grid; pause

figure; % only second half of the signal without initial transient
nn = Nx/2+1:Nx; f0=fs/(Nx/2); f=f0*(0:Nx/2-1); w=chebwin(Nx/2,100)'; w=w/sum(w);
subplot(211); plot(f,20*log10(abs(fft(w.*x(nn))))); grid; xlabel('f [Hz]'); title('IN  |X(f)| [dB]');
subplot(212); plot(f,20*log10(abs(fft(w.*y(nn))))); grid; xlabel('f [Hz]'); title('OUT |Y(f)| [dB]');
pause
